@@ -7,6 +7,7 @@ import re
 import weakref
 from copy import deepcopy
 from lxml import etree
+from enum import Enum, unique
 
 HERE = os.path.abspath(os.path.dirname(__file__))
 XSD_PATH = os.path.join(HERE, 'resources', 'SCL_Schema', 'SCL.xsd')
@@ -17,7 +18,8 @@ MAX_VALIDATION_SIZE = 500  # taille en Mo
 REG_DA = r'(?:\{.+\})?[BS]?DA'
 REG_DO = r'(?:\{.+\})?S?DO'
 REG_SDI = r'(?:\{.+\})?S?D[OA]?I'
-REG_ARRAY_TAGS = r'(?:\{.+\})?(?:FCDA|ClientLN|IEDName|FIP|BAP|ExtRef|Terminal|P|DataSet)'  # |Server)'
+REG_ARRAY_TAGS = r'(?:\{.+\})?(?:FCDA|ClientLN|IEDName|FIP|BAP|ExtRef|Terminal|P|DataSet|GSE' \
+                 r'|GSEControl|ReportControl|VoltageLevel)'  # |Server)'
 REG_DT_NODE = r'(?:.*\})?((?:[BS]?D[AO])|(?:LN0?))'
 REF_SCL_NODES = r'(?:\{.+\})?(?:Header|Substation|Private|Communication)'
 DEFAULT_AP = 'PROCESS_AP'
@@ -107,9 +109,20 @@ NODES_ATTRS = {
         'fc',
         'ix'
     ],
+    'DataSet': [
+        'name'
+    ]
 }
 
 
+@unique
+class ServiceType(str, Enum):
+    Poll = 'Poll'
+    Report = 'Report'
+    GOOSE = 'GOOSE'
+    SMV = 'SMV'
+
+
 def _safe_convert_value(value: str) -> any:
     """
         Convert a string value in typed value une valeur string en valeur typée.
@@ -146,9 +159,29 @@ def _safe_convert_value(value: str) -> any:
         return None
 
 
+def _get_tag_without_ns(nstag: str) -> str:
+    """
+        Get the xml tag without namespace
+
+        Parameters
+        ----------
+        nstag
+           The xml element tag
+
+        Returns
+        -------
+        str
+            the tag without namespace
+    """
+    tag_reg = r'(?:{.+})?(\w+)'
+    result = re.match(tag_reg, nstag)
+    if result:
+        return result.group(1)
+
+
 def _get_node_name(node: etree.Element):
     name = None
-    tag = node.tag.split('}')[-1]
+    tag = _get_tag_without_ns(node.tag)
 
     name = node.get('name')
     if not name:
@@ -185,7 +218,7 @@ class DataTypeTemplates:
                 xml_path
                     chemin du fichier SCD/SCL contenant les datatypes
         """
-        context = etree.iterparse(xml_path, events=("end",), tag='{}DataTypeTemplates'.format(SCL_NAMESPACE))
+        context = etree.iterparse(xml_path, events=("end",), tag='{}DataTypeTemplates'.format(SCL_NAMESPACE), remove_comments=True)
         _, self._datatypes_root = next(context)
 
     def get_type_by_id(self, id: str) -> etree.Element:
@@ -227,10 +260,6 @@ class DataTypeTemplates:
         return tags
 
 
-# TODO :
-# - Manage BAP/FIP
-# - Gestion des attributs dont le nom comporte des . ou des -
-# - Gestion des noeud étranges (voir SCD de Gilles)
 class SCDNode:
     """
         Basic class to compute SCD nodes
@@ -285,7 +314,7 @@ class SCDNode:
         self._fullattrs = fullattrs
 
         if node_elem is not None:
-            self.tag = node_elem.tag.split('}')[-1]
+            self.tag = _get_tag_without_ns(node_elem.tag)
 
         if node_elem is None and not kwargs:
             msg = 'Please enter a xml Element and/or attributes (kwargs)'
@@ -319,7 +348,7 @@ class SCDNode:
         new_node = None
         attributes = {}
         attributes.update(elem.attrib)
-        _tag = elem.tag.split('}')[-1]
+        _tag = _get_tag_without_ns(elem.tag)
 
         if elem.text:
             val = elem.text.strip()
@@ -333,12 +362,14 @@ class SCDNode:
             new_node = DA(self._datatypes, elem, self._fullattrs, **attributes)
         elif re.fullmatch(REG_DO, elem.tag):
             new_node = DO(self._datatypes, elem, self._fullattrs, **attributes)
-        elif elem.tag.split('}')[-1] == 'LN':
+        elif _get_tag_without_ns(elem.tag) == 'LN':
             new_node = LN(self._datatypes, elem, self._fullattrs, **attributes)
-        elif elem.tag.split('}')[-1] == 'LN0':
+        elif _get_tag_without_ns(elem.tag) == 'LN0':
             new_node = LN0(self._datatypes, elem, self._fullattrs, **attributes)
-        elif elem.tag.split('}')[-1] == 'LDevice':
+        elif _get_tag_without_ns(elem.tag) == 'LDevice':
             new_node = LD(self._datatypes, elem, self._fullattrs, **attributes)
+        elif _get_tag_without_ns(elem.tag) == 'DataSet':
+            new_node = DataSet(self._datatypes, elem, self._fullattrs, **attributes)
         else:
             new_node = SCDNode(self._datatypes, elem, self._fullattrs, **attributes)
 
@@ -360,17 +391,36 @@ class SCDNode:
             Returns
             -------
             `array`
-                Return an dictionnary of found children SCDNodes : array[IntAdr] = SCDNode
-                (IntAdr is the path build with the names of the ancestors starting at the SCDNode)
+                Return an dictionnary of found children SCDNodes : int_addr: SCDNode
+                (int_addr is the path build with the names of the ancestors starting at the SCDNode)
         """
 
         leaves = {}
-        mms = False
-        if self.tag in ['IED', 'LN', 'LN0', 'LDevice']:
-            mms = True
-        self._collect_DA_leaf_nodes(self, leaves, mms)
+        self._collect_DA_leaf_nodes(self, leaves)
         return leaves
 
+    def get_DO_nodes(self) -> dict:
+        """
+            Recursively retrieve the leaf DO of the current node.
+
+            Returns
+            -------
+            `array`
+                Return an dictionnary of found children SCDNodes : int_addr: SCDNode
+                (int_addr is the path build with the names of the ancestors starting at the SCDNode)
+        """
+        do_nodes = {}
+        if isinstance(self, (DA, DO)):
+            node = self
+            while node.parent is not None:
+                if isinstance(node.parent(), LN):
+                    do_nodes[self.get_int_addr(node)] = node
+                node = node.parent()
+        else:
+            self._collect_DO_nodes(self, do_nodes)
+
+        return do_nodes
+
     def get_children(self, tag: str = None) -> list:
         """
             Retrieve the children nodes of the current node.
@@ -388,14 +438,58 @@ class SCDNode:
         """
 
         results = []
-        for _, item in self.__dict__.items():
+        for key, item in self.__dict__.items():
             if isinstance(item, SCDNode):
                 if tag is None:
                     results.append(item)
                 elif item.tag == tag:
                     results.append(item)
+            elif isinstance(item, list) and re.fullmatch(REG_ARRAY_TAGS, key):
+                if tag is None:
+                    results += item
+                elif key == tag:
+                    results += item
         return results
 
+    def get_name_subtree(self, fc_filter: str = None):
+        """
+            get_name_subtree
+
+            Parameters
+            ----------
+            `fc_filter` (optional)
+                Functional constraint string
+
+            Return
+            ----------
+            `i_obj_path`
+                Tree of 2-tuple elements (name, [subelem])
+        """
+        tree = {}
+        node_depth = len(self.get_int_addr(self).split("."))
+
+        for leaf_path, leaf_node in self.get_DA_leaf_nodes().items():
+            if fc_filter is None or leaf_node.get_associated_fc() == fc_filter:
+                node = tree
+                for level in leaf_path.split(".")[node_depth-1:]:
+                    node = node.setdefault(level, dict())
+        tree_as_tuples = self._recursive_dict_to_tuple(tree)
+        return tree_as_tuples[0]
+
+    def _recursive_dict_to_tuple(self, i_dict):
+        """
+            /!\\ PRIVATE : do not use /!\\
+
+            Recursive reshaping dict tree in list of
+            2-tuple: ('name of the node', list of 2-tuple children)
+
+            Parameters
+            ----------
+            `i_dict` (optional)
+                dict to reshape
+        """
+        return [(k, self._recursive_dict_to_tuple(v)) for k, v in i_dict.items()] if i_dict else []
+
     def _create_node(self, node_elem: etree.Element, **kwargs):
         """
             /!\\ PRIVATE : do not use /!\\
@@ -469,7 +563,7 @@ class SCDNode:
 
         return len(node.get_children()) == 0 and isinstance(node, DA) and hasattr(node, 'parent')
 
-    def _collect_DA_leaf_nodes(self, node, leaves: dict, mms: bool = False) -> dict:
+    def _collect_DA_leaf_nodes(self, node, leaves: dict) -> dict:
         """
             /!\\ PRIVATE : do not use /!\\
 
@@ -483,41 +577,138 @@ class SCDNode:
             `leaves`
                 The found leaves dictionnary
 
-            `mms`
-                If True, compute the mms and u_mms adresses
-
             Returns
             -------
-            `{IntAdr : node}`
+            `{int_addr : node}`
                 The found leaves dictionnary.
         """
         if node is not None:
             if self._is_leaf(node):
-                ancestor = node.parent()
-                IntAdr = node.name
-                while ancestor is not None:
-                    IntAdr = ancestor.name + '.' + IntAdr
-                    if ancestor.parent is not None and ancestor.tag != 'LDevice':
-                        ancestor = ancestor.parent()
-                    else:
-                        ancestor = None
+                leaves[self.get_int_addr(node)] = node
 
-                mmsAdr = ''
-                u_mmsAdr = ''
-
-                if mms:  # mms adresses ok only if the root is a LD
-                    for item in IntAdr.split('.'):
-                        mmsAdr += SEP1 + item
-                        u_mmsAdr += SEP2 + item
-                    setattr(node, 'mmsAdr', mmsAdr[1:])
-                    setattr(node, 'u_mmsAdr', u_mmsAdr[1:])
-
-                setattr(node, 'IntAdr', IntAdr)
-                leaves[IntAdr] = node
             else:
                 children = node.get_children()
                 for child in children:
-                    self._collect_DA_leaf_nodes(child, leaves, mms)
+                    self._collect_DA_leaf_nodes(child, leaves)
+
+    def _collect_DO_nodes(self, node, do_nodes: dict) -> dict:
+        """
+            /!\\ PRIVATE : do not use /!\\
+
+            Retrieve DO nodes
+
+            Parameters
+            ----------
+            `node`
+                SCDNode to compute
+
+            `do`
+                The found do
+
+            Returns
+            -------
+            `{int_addr : node}`
+                The found leaves dictionnary.
+        """
+        if node is not None:
+            if isinstance(node, DO):
+                do_nodes[self.get_int_addr(node)] = node
+
+            else:
+                children = node.get_children()
+                for child in children:
+                    self._collect_DO_nodes(child, do_nodes)
+
+
+    def get_int_addr(self, node) -> str:
+        """
+            Get int adr of SCDNode
+
+            Parameters
+            ----------
+            `node`
+                The SCDNode to check
+
+            Returns
+            -------
+            `str`
+                Return the internal address of the node
+        """
+        assert isinstance(node, (LD, LN, DO, DA)), "Invalid SCDNode level, expect LD, LN, DO or DA"
+        int_addr = node.name
+        if node.parent is None:
+            logging.debug(f'SCDNode::get_int_addr: parent node of node: {self.name} is None, cannot build int_addr')
+            return None
+        ancestor = node.parent()
+        while ancestor is not None:
+            int_addr = ancestor.name + '.' + int_addr
+            if ancestor.parent is not None and ancestor.tag != 'LDevice':
+                ancestor = ancestor.parent()
+            else:
+                ancestor = None
+        return int_addr
+
+    def get_object_reference(self) -> str:
+        """
+            Get object reference as defined by IEC61850-8-1, section 8.1.3.2.3
+
+            Returns
+            -------
+            `str`
+                Return the object reference of the node
+        """
+        assert isinstance(self, (LD, LN, DO, DA, DataSet)), "Invalid SCDNode level, expects LD, LN, DO, DA or DataSet"
+        if isinstance(self, LD):
+            if hasattr(self, 'ldName'):
+                return self.ldName
+            else:  # ldName is optional, if absent, use <IEDName><LDInst>
+                if self.parent is None:
+                    logging.debug(f'SCDNode::_get_object_reference: parent object not found for LD {self.name}')
+                    return None
+                return self.parent().parent().name + self.name
+
+        else:
+            if self.parent is None:
+                logging.debug(f'SCDNode::_get_object_reference: parent node not found '
+                              f'to build object reference for node with name {self.name}')
+                return None
+            sep = '/' if isinstance(self.parent(), LD) else '.'
+            return self.parent().get_object_reference() + sep + self.name
+
+    def get_GOCB_reference(self) -> str:
+        """
+            Get GOOSE Control block Reference, with format ldName/lnName$GO$CBName
+                NOTE: If a class GSEControl is created, this should belong to it
+
+            Returns
+            -------
+            `str`
+                Return the reference of the GOOSE Control Block
+        """
+        assert self.tag == 'GSEControl', "Invalid SCDNode, expects tag GSEControl"
+        ied = self.get_parent_with_class(IED)
+        ld = self.get_parent_with_class(LD)
+        ln = self.get_parent_with_class(LN)  # expected is LLN0
+        if ied and ld and ln:
+            return f'{ied.name}{ld.name}/{ln.name}$GO${self.name}'
+
+
+    def get_parent_with_class(self, parent_class: type):
+        """
+            Get parent node with input Class (e.g. IED, LD, LN, DO, DA)
+
+            Returns
+            -------
+            `SCDNode`
+                Return the first node matching request, else None
+        """
+        p = self
+        while True:
+            if p.parent is None:
+                return None
+            p = p.parent()
+            if isinstance(p, parent_class):
+                return p
 
     def _create_by_node_elem(self, node: etree.Element):
         """
@@ -598,7 +789,7 @@ class SCDNode:
 
         for elem in children:
             if not elem.get('id'):  # No instances in datatypes
-                tag = elem.tag.split('}')[-1]
+                tag = _get_tag_without_ns(elem.tag)
                 if tag == 'Val' and elem.text:
                     setattr(self, 'Val', _safe_convert_value(elem.text))
                 elif re.fullmatch(REG_SDI, tag):
@@ -628,7 +819,7 @@ class SCDNode:
         if hasattr(current_node, inst_node.get('name')):
             upd_node = getattr(current_node, inst_node.get('name'))
             if isinstance(upd_node, SCDNode):
-                if inst_node.tag.split('}')[-1] == 'DAI':
+                if _get_tag_without_ns(inst_node.tag) == 'DAI':
                     upd_node._set_attributes_from_elem(inst_node)
 
                 upd_node._set_instances(inst_node)
@@ -662,6 +853,41 @@ class DA(SCDNode):
         self._all_attributes = NODES_ATTRS['DA']
         super().__init__(datatypes, node_elem, fullattrs, **kwargs)
 
+    def get_associated_fc(self):
+        """
+            Returns the Functional Constraint associated with DA. If absent (SDA case), climb the hierarchy
+
+            Returns
+            -------
+            `fc`
+                Return an the value of the 'fc' field of the DA
+        """
+        if hasattr(self, "fc"):
+            return self.fc
+        else:
+            if isinstance(self.parent(), DA):
+                return self.parent().get_associated_fc()
+            else:
+                err = f'SCD not valid: No FC found for DA "{self.name}"'
+                LOGGER.error(err)
+                raise AttributeError(err)
+
+    def get_mms_var_name(self) -> str:
+        """
+            Get MMS variable name as defined by IEC61850-8-1, section 7.3.4 (including domain name)
+
+            Returns
+            -------
+            `str`
+                Return the MMS variable name of the node
+        """
+        obj_ref = self.get_object_reference()
+        if not obj_ref:
+            raise AttributeError('DO::get_mms_var_name: invalid object reference')
+        obj_ref_parts = obj_ref.split(".")
+        obj_ref_parts.insert(1, self.get_associated_fc())
+        return "$".join(obj_ref_parts)
+
 
 class DO(SCDNode):
     """
@@ -691,6 +917,26 @@ class DO(SCDNode):
         self._all_attributes = NODES_ATTRS['DO']
         super().__init__(datatypes, node_elem, fullattrs, **kwargs)
 
+    def get_mms_var_name(self, fc: str) -> str:
+        """
+            Get MMS variable name as defined by IEC61850-8-1, section 7.3.3 (including domain name)
+
+            Parameter
+            -------
+            `fc`
+                Functional constraint string to include in the variable name
+
+            Returns
+            -------
+            `str`
+                Return the MMS variable name of the node
+        """
+        obj_ref = self.get_object_reference()
+        if not obj_ref:
+            raise AttributeError('DO::get_mms_var_name: invalid object reference')
+        obj_ref_parts = obj_ref.split(".")
+        obj_ref_parts[0] += '$' + fc
+        return "$".join(obj_ref_parts)
 
 class LN(SCDNode):
     """
@@ -764,6 +1010,127 @@ class LN0(LN):
         super().__init__(datatypes, node_elem, fullattrs, **kwargs)
 
 
+class DataSet(SCDNode):
+    """
+        Class to manage a DataSet
+    """
+    _data_tree = None
+
+    def __init__(self, datatypes: DataTypeTemplates, node_elem: etree.Element = None, fullattrs: bool = False, **kwargs: dict):
+        """
+            Constructor
+
+            Parameters
+            ----------
+            `datatypes`
+                Instance of the DataTypeTemplates object from the SCD/SCL file.
+
+            `node_elem` (optional)
+                etree.Element element from the SCD/SCL file to build the node object.
+
+            `fullattrs` (optional)
+                If True, all the possible attributes for the SCD objects will be created
+                even if they are not described in the SCD/SCL file.
+
+            `kwargs` (optional)
+                Dictionary of the node attributes.
+
+            /!\\ At least one of node_elem or kwargs must be provided /!\\
+        """
+        super().__init__(datatypes, node_elem, fullattrs, **kwargs)
+        self._all_attributes = NODES_ATTRS['DataSet']
+
+    def as_tree(self) -> dict:
+        """
+            Get the DataSet elements as a tree matching the structure of data sent in control blocks
+            Root node of Dataset is named 'root', e.g.:
+                 ('root', [
+                    ('pathuntilSDA1', [
+                        ('DA1', []),
+                        ('DA2', [
+                            ('SDA21', []),
+                            ('SDA22', [])
+                        ])
+                    ]),
+                    ('pathuntilSDA2', [
+                       ('DA1', []),
+                       ('DA2', [])
+                    ])
+                ])
+
+            Returns
+            -------
+            `()`
+                DataSet element as a tree of DO and DA.
+        """
+        if self._data_tree is not None:
+            return self._data_tree
+
+        self._data_tree = ('root', [])  # tree starting with root, then object ref until node, then subsequent nodes
+        ld = self.parent().parent()  # DataSet is located in LD.LLN0
+
+        for fcda in self.FCDA:
+            fcda_fc = fcda.fc
+            fcda_doName = fcda.doName
+            fcda_prefix = fcda.prefix if hasattr(fcda, "prefix") else ""
+            fcda_daName = fcda.daName if hasattr(fcda, "daName") else ""
+            fcda_lnClass = fcda.lnClass
+            fcda_lnInst = "" if fcda_lnClass == "LLN0" else fcda.lnInst
+            fcda_ln_full_name = fcda_prefix + fcda_lnClass + str(fcda_lnInst)
+            fcda_dataref = f"{ld.name}.{fcda_ln_full_name}.{fcda_doName}{'.' + fcda_daName if fcda_daName else ''}"
+
+            path_to_fcda_node = fcda_dataref.split(".")
+            node = ld.get_LN_by_name(fcda_ln_full_name)
+            for i in range(1, len(path_to_fcda_node)):
+                if hasattr(node, path_to_fcda_node[i]):
+                    node = getattr(node, path_to_fcda_node[i])
+
+            self._data_tree[1].append((fcda_dataref, node.get_name_subtree(fc_filter=fcda_fc)[1]))
+
+        return self._data_tree
+
+    def as_list(self) -> list:
+        """
+            Get the array of DA path present in a dataset
+
+            Returns
+            -------
+            `[]`
+                List of DA in DataSet
+        """
+        dataset_path_list = self._tree_to_list(self.as_tree())
+        return [".".join(da_path[1:]) for da_path in dataset_path_list]  # skip 'root' element
+
+    def get_path_in_dataset(self, da_path: str) -> list:
+        """
+            Convert a DA path into a list of keys corresponding to tree nodes
+
+            Parameters
+            ----------
+            `da_path`
+                DA path (LD.LN.DO[.SDO].DA[.SDA])
+
+            Returns
+            -------
+            `[]`
+                list of keys to reach DA in Dataset tree
+        """
+        seq = [path_list for path_list in self._tree_to_list(self.as_tree()) if ".".join(path_list[1:]) == da_path]
+        if len(seq) == 0:
+            raise AttributeError(f"get_path_in_dataset: DA '{da_path}' not found in DataSet '{self.name}'")
+        return seq[0]
+
+    def _tree_to_list(self, n):
+        if len(n[1]):
+            l = []
+            for next_n in n[1]:
+                for e in self._tree_to_list(next_n):
+                    l.append([n[0]] + e)
+            return l
+        else:
+            return [[n[0]]]
+
+
 class LD(SCDNode):
     """
         Class to manage a LD
@@ -797,19 +1164,19 @@ class LD(SCDNode):
         else:
             inst = kwargs.get('inst')
             kwargs.update({'name': inst})
-
         super().__init__(datatypes, node_elem, fullattrs, **kwargs)
 
-    def get_inputs_goose_extrefs(self) -> list:
+    def get_inputs_extrefs(self, service_type: ServiceType = None) -> list:
         """
-            Get the input GOOSE extRefs list
+            Get the input extRefs list
 
             Returns
             -------
             `[]`
                 An array of objects containing the extRefs attributes
         """
-        XPATH_INPUTS_EXTREFS = './iec61850:LN0/iec61850:Inputs/iec61850:ExtRef[@serviceType="GOOSE"]'
+        service_type_str = f'[@serviceType="{service_type.value}"]' if service_type else ''
+        XPATH_INPUTS_EXTREFS = f'./iec61850:LN0/iec61850:Inputs/iec61850:ExtRef{service_type_str}'
         extrefs = []
 
         xpath_result = self._node_elem.xpath(XPATH_INPUTS_EXTREFS, namespaces=NS)
@@ -820,11 +1187,93 @@ class LD(SCDNode):
 
         return extrefs
 
+    def get_goose_inputs_extrefs(self) -> list:
+        """
+            Get the GOOSE input extRefs list
+
+            Returns
+            -------
+            `[]`
+                An array of objects containing the GOOSE extRefs attributes
+        """
+        return self.get_inputs_extrefs(ServiceType.GOOSE)
+
+    def get_datasets(self) -> list:
+        """
+            Get the GSEControl list
+
+            Returns
+            -------
+            `[]`
+                An array of objects containing the GSEControl attributes
+        """
+        return self.LLN0.DataSet if hasattr(self.LLN0, "DataSet") else []
+
+    def get_dataset_by_name(self, name: str) -> DataSet:
+        """
+            Get the DataSet
+
+            Returns
+            -------
+            `node`
+                GSEControl with input name, None if not found
+        """
+        filtered_dataset = [d for d in self.get_datasets() if d.name == name]
+        return filtered_dataset[0] if len(filtered_dataset) == 1 else None
+
+    def get_gsecontrols(self) -> list:
+        """
+            Get the GSEControl list
+
+            Returns
+            -------
+            `[]`
+                An array of objects containing the GSEControl attributes
+        """
+        return self.LLN0.GSEControl if hasattr(self.LLN0, "GSEControl") else []
+
+    def get_gsecontrol_by_name(self, name: str) -> SCDNode:
+        """
+            Get the GSEControl
+
+            Returns
+            -------
+            `node`
+                GSEControl with input name, None if not found
+        """
+        filtered_gsecontrol = [g for g in self.get_gsecontrols() if g.name == name]
+        return filtered_gsecontrol[0] if len(filtered_gsecontrol) == 1 else None
+
+    def get_reportcontrols(self) -> list:
+        """
+            Get the ReportControl list
+
+            Returns
+            -------
+            `[]`
+                An array of objects containing the ReportControl attributes
+        """
+        return self.LLN0.ReportControl if hasattr(self.LLN0, "ReportControl") else []
+
+    def get_reportcontrol_by_name(self, name: str) -> SCDNode:
+        """
+            Get the ReportControl
+
+            Returns
+            -------
+            `node``
+                ReportControl with input name, None if not found
+        """
+        filtered_reportcontrol = [r for r in self.get_reportcontrols() if r.name == name]
+        return filtered_reportcontrol[0] if len(filtered_reportcontrol) == 1 else None
+
     def get_LN_by_name(self, ln_Name: str) -> LN:
         if hasattr(self, ln_Name):
             return getattr(self, ln_Name)
 
 
+
+
 class IED(SCDNode):
     """
         Class to manage an IED
@@ -857,25 +1306,27 @@ class IED(SCDNode):
         self._LDs = {}
         super().__init__(datatypes, node_elem, fullattrs, **kwargs)
 
-    def get_inputs_goose_extrefs(self) -> list:
+    def get_inputs_extrefs(self, service_type: ServiceType = None) -> list:
         """
-            Get the input GOOSE extRefs list
+            Get the input extRefs list
 
             Returns
             -------
             `[]`
                 An array of objects containing the extRefs attributes
         """
-        XPATH_INPUTS_EXTREFS = './iec61850:AccessPoint[@name="PROCESS_AP"]/*/iec61850:LDevice/iec61850:LN0/iec61850:Inputs/iec61850:ExtRef'  # [@serviceType="GOOSE"]
-        extrefs = []
+        return [e for ld in self.get_children_LDs() for e in ld.get_inputs_extrefs(service_type)]
 
-        xpath_result = self._node_elem.xpath(XPATH_INPUTS_EXTREFS, namespaces=NS)
+    def get_goose_inputs_extrefs(self) -> list:
+        """
+            Get the GOOSE input extRefs list
 
-        for item in xpath_result:
-            itm = deepcopy(item.attrib)
-            extrefs.append(itm)
-
-        return extrefs
+            Returns
+            -------
+            `[]`
+                An array of objects containing the GOOSE extRefs attributes
+        """
+        return self.get_inputs_extrefs(ServiceType.GOOSE)
 
     def get_children_LDs(self, ap_name: str = DEFAULT_AP) -> list:
         ap = self._get_ap_by_name(ap_name)
@@ -897,11 +1348,45 @@ class IED(SCDNode):
         ln = ld.get_LN_by_name(ln_Name)
         return ln
 
+    def get_node_by_path(self, data_ref: str, ap_name: str = DEFAULT_AP) -> SCDNode:
+        f"""
+            Get a SCDNode from its reference
+
+            Parameters
+            ----------
+            `data_ref`
+                reference in the format 'LD.LN.DO[.SDO(s)].DA[.SDA(s)]'
+
+            `ap_name` (optional)
+                Access point name (default = {DEFAULT_AP})
+
+            Returns
+            -------
+            `node`
+                SCDNode
+        """
+        data_path = data_ref.replace("/", ".").split(".")
+        node = self
+        if len(data_path) > 0 and data_path[0] == self.name:
+            data_path.pop(0)  # remove IED name if present
+        if len(data_path) > 0:
+            node = self.get_LD_by_inst(data_path.pop(0), ap_name)
+        if len(data_path) > 0:
+            node = node.get_LN_by_name(data_path.pop(0))
+        while len(data_path) > 0:
+            node = getattr(node, data_path.pop(0))
+        if node is None:
+            err = f'Could not reach data node with reference {data_ref} in IED {self.name}'
+            LOGGER.error(err)
+            raise AttributeError(err)
+        return node
+
     def _get_ap_by_name(self, ap_name):
         if hasattr(self, ap_name):
             return getattr(self, ap_name)
 
 
+
 class SCD_handler():
     """
         Class to handle a SCD/SCL file
@@ -936,7 +1421,7 @@ class SCD_handler():
         self.Substation = []
         scl_children = self._get_SCL_elems()
         for elem in scl_children:
-            if elem.tag.split('}')[-1] == 'Substation':
+            if _get_tag_without_ns(elem.tag) == 'Substation':
                 self.Substation.append(SCDNode(self.datatypes, elem, self._fullattrs))
             else:
                 elem_name = _get_node_name(elem)
@@ -960,7 +1445,7 @@ class SCD_handler():
         path, ext = os.path.splitext(self._scd_path)
         dest_path = '{}-reduced{}'.format(path, ext)
 
-        ctx = etree.iterparse(self._scd_path, events=("start",), tag='{}SCL'.format(SCL_NAMESPACE))
+        ctx = etree.iterparse(self._scd_path, events=("start",), tag='{}SCL'.format(SCL_NAMESPACE), remove_comments=True)
         for _, root in ctx:
             newroot = etree.Element(root.tag, nsmap=root.nsmap, attrib=root.attrib)
             break
@@ -1020,7 +1505,7 @@ class SCD_handler():
             `IED`
                 The loaded IED object
         """
-        if hasattr(self._IEDs, ied_name):
+        if ied_name in self._IEDs:
             return self._IEDs[ied_name]
 
         ied_elems = self._get_IED_elems_by_names([ied_name])
@@ -1078,6 +1563,14 @@ class SCD_handler():
         return result
 
     def get_IP_Adr(self, ied_name: str):
+        """
+            Get IP address from GSE for an IED
+
+            Returns
+            -------
+            `str`
+                IP of the IED
+        """
         for iComm in self.Communication.get_children('SubNetwork'):  # browse all iED SubNetWork
             if iComm.type != "8-MMS":  # IP can be found only in MMS access point '.
                 continue
@@ -1091,6 +1584,23 @@ class SCD_handler():
                                     return iP.Val, iCnxAP.apName
             return None, None    # Not found
 
+    def get_GSEs(self, ied_name: str) -> list:
+        """
+            List the GSE SCDNodes of the IED
+
+            Returns
+            -------
+            `[]`
+                List of GSEs
+        """
+        for subnet in self.Communication.get_children('SubNetwork'):  # browse all iED SubNetWork
+            if subnet.type != "8-MMS":
+                continue
+            for conn_ap in subnet.get_children('ConnectedAP'):  # browse all Access Point(s) of the iED
+                if conn_ap.iedName == ied_name:
+                    return conn_ap.GSE if hasattr(conn_ap, "GSE") else []
+        return []
+
     def _check_scd_file(self) -> tuple:
         """
             /!\\ PRIVATE : do not use /!\\
@@ -1109,7 +1619,7 @@ class SCD_handler():
         """
         file_size = os.path.getsize(self._scd_path) // (1024 * 1024)
         if file_size < MAX_VALIDATION_SIZE:
-            tree = etree.parse(self._scd_path)
+            tree = etree.parse(self._scd_path, parser=etree.XMLParser(remove_comments=True))
             is_valid = self._schema.validate(tree)
             if not FORCE_ITER_MODE:
                 self._scl_root = tree.getroot()
@@ -1236,7 +1746,7 @@ class SCD_handler():
             `[etree.Element]`
                 Array of found etree.Element elements
         """
-        context = etree.iterparse(self._scd_path, events=("end",), tag='{}{}'.format(SCL_NAMESPACE, tag))
+        context = etree.iterparse(self._scd_path, events=("end",), tag='{}{}'.format(SCL_NAMESPACE, tag), remove_comments=True)
 
         elem_list = []
         _, elem = next(context)
@@ -1266,15 +1776,15 @@ class SCD_handler():
         for idx, tag in enumerate(tags):
             tags[idx] = '{}{}'.format(SCL_NAMESPACE, tag)
 
-        context = etree.iterparse(self._scd_path, events=("end",), tag=tags)
+        context = etree.iterparse(self._scd_path, events=("end",), tag=tags, remove_comments=True)
         result = []
         for _, elem in context:
 
-            if elem.tag.split('}')[-1] != 'Private':
+            if _get_tag_without_ns(elem.tag) != 'Private':
                 result.append(elem)
-            elif elem.tag.split('}')[-1] == 'Private' \
-                                            and (elem.xpath('following-sibling::iec61850:Header', namespaces=NS) or
-                                                 elem.xpath('preceding-sibling::iec61850:Header', namespaces=NS)):
+            elif _get_tag_without_ns(elem.tag) == 'Private' and \
+                    (elem.xpath('following-sibling::iec61850:Header', namespaces=NS) or
+                     elem.xpath('preceding-sibling::iec61850:Header', namespaces=NS)):
                 result.append(elem)
             else:
                 elem.clear()
@@ -1297,7 +1807,7 @@ class SCD_handler():
             `[etree.Element]`
                 Array of found etree.Element elements
         """
-        context = etree.iterparse(self._scd_path, events=("end",), tag='{}IED'.format(SCL_NAMESPACE))
+        context = etree.iterparse(self._scd_path, events=("end",), tag='{}IED'.format(SCL_NAMESPACE), remove_comments=True)
         result = []
         for _, ied in context:
             item_name = ied.get('name')
@@ -1324,7 +1834,7 @@ class SCD_handler():
             `[etree.Element]`
                 Array of found etree.Element elements
         """
-        context = etree.iterparse(self._scd_path, events=("end",), tag='{}IED'.format(SCL_NAMESPACE))
+        context = etree.iterparse(self._scd_path, events=("end",), tag='{}IED'.format(SCL_NAMESPACE), remove_comments=True)
         result = []
         for _, ied in context:
             item_type = ied.get('type')
@@ -1345,9 +1855,10 @@ class SCD_handler():
                 The found IED names
         """
         result = []
-        context = etree.iterparse(self._scd_path, events=("end",), tag='{}IED'.format(SCL_NAMESPACE))
+        context = etree.iterparse(self._scd_path, events=("end",), tag='{}IED'.format(SCL_NAMESPACE), remove_comments=True)
         for _, ied in context:
             result.append(ied.get('name'))
             ied.clear()
 
         return result
+
